#include "myfndt.h"
#include "dynstem.h"
#include "stemimpl.h"

#define PROTO(rtype, name, par)		typedef rtype (*F##name)par;
#include "libstem.h"

#include "locusopt.h"
#include "odlerr.h"
#include "ouno.h"
#include "ometastm.h"
#include <dlfcn.h>

CDynamicStemmer::CDynamicStemmer(CLocusOptions &Options):
    Lib(NULL),
    Stemmer(NULL),
    ApplyProc(NULL),
    DoneProc(NULL),
    DlsymErrMsgFormat(Options.Get(T_DLSYMERR_SS))
{    
    try
    {   InitLib(GetLibraryName(Options.Get(O_STEMLIB)),
	    Options.Get(T_DLOPENERR_SS));
        FGeneric InitProc = GetFncAddr("StemInit");
        Stemmer = (reinterpret_cast<FStemInit>(InitProc))(
	    Options.Get(O_UNOFILE));
        ApplyProc = GetFncAddr("StemApply");
        DoneProc = GetFncAddr("StemDone");
    }
    catch (...)
    {   Done();
        throw;
    }
}

DString CDynamicStemmer::GetLibraryName(const DString &StemImpl)
{
    for (int i = 0;
	i < sizeof(StemImplementations) / sizeof(StemImplementations[0]);
	++i)
    {   if (StemImpl == StemImplementations[i])
        {   CMemPtr<char> LibName(StemImpl.GetLength() + 20);
	    sprintf((char *)LibName, "%s.so.%d",
		StemImpl.c_str(),
		StemLibraries[i]);
	    return DString((char *)LibName);
	}
		
    }
    Assert(0);
    return DString("");
}

CDynamicStemmer::~CDynamicStemmer()
{
    Done();
}

void CDynamicStemmer::Done()
{
    if (Lib)
    {   if (DoneProc && Stemmer)
            (reinterpret_cast<FStemDone>(DoneProc))(Stemmer);
        dlclose(Lib);
    }
}

int CDynamicStemmer::Apply(DString &WordForm)
{
    return (reinterpret_cast<FStemApply>(ApplyProc))(Stemmer, WordForm);
}

void CDynamicStemmer::InitLib(const DString &LibName,
    const DString &ErrMsgFormat)
{   
    Lib = dlopen(LibName.c_str(), RTLD_LAZY | RTLD_GLOBAL);
    if (!Lib)
    {   const char *e = dlerror();
        DString err = e ? e : "";
        CMemPtr<char> Msg(ErrMsgFormat.GetLength() +
            LibName.GetLength() +
            err.GetLength() +
            10);
        sprintf((char *)Msg,
	    ErrMsgFormat.c_str(),
	    LibName.c_str(),
	    err.c_str());
        throw CGeneralException((char *)Msg);
    }
}

CDynamicStemmer::FGeneric CDynamicStemmer::GetFncAddr(const char *FncName)
{
    void *Addr = dlsym(Lib, FncName);
    if (!Addr)
    {   const char *e = dlerror();
    DString err = e ? e : "";
        CMemPtr<char> Msg(DlsymErrMsgFormat.GetLength() +
            strlen(FncName) +
            err.GetLength() +
            10);
        sprintf((char *)Msg, DlsymErrMsgFormat.c_str(), FncName, err.c_str());
        throw CGeneralException((char *)Msg);
    }
    return reinterpret_cast<FGeneric>(Addr);
}
