#if !defined(__univopt_h)
#define __univopt_h

#include "opt1.h"
#include "opt2.h"
#include "opt3.h"
#include "opt4.h"

#include <vector.h>

// templates are split into multiple modules to keep segment
// under 64K
class CUnivOptions : public COptions1,		     
    public COptions2,
    public COptions3,
    public COptions4
{
private:
    vector<DString> SectList; // sorted
    char *LastErrorMessage;
protected:
    int ComposedKeys;

public:
    CUnivOptions();
    virtual ~CUnivOptions();

    // returns empty string OK, otherwise key which has
    // initial invalid value
    virtual DString FirstInvalidKey();

    void WatchSection(const char *SectName);
    void ChangeWatchedSections(const CStringListValue &ToList);

    // bug: can throw and caller does not handle it
    virtual short HasSection(const char *SectName);

    inline void Add(const char *Key, const char *Value);
    inline void Add(const char *Key, const DString &Value);
    inline void Add(const char *Key, COptions<DString>::FCompute Compute);
    inline void Add(const char *Key, const CStringListValue &Value);
    inline void Add(const char *Key,
       	COptions<CStringListValue>::FCompute Compute);
    inline void Add(const char *Key, int Value);
	// no need to pass int by reference
    inline void Add(const char *Key, COptions<int>::FCompute Compute);
    inline void Add(const char *Key, double &Value);	
    inline void Add(const char *Key, COptions<double>::FCompute Compute);

    // dtto
    inline short Override(const char *Section, 
	const char *Key, 
	const DString &Value);
    inline short Override(const char *Section,
	const char *Key, 
	const CStringListValue &Value);
    inline short Override(const char *Section,
	const char *Key, 
	int Value);
    inline short Override(const char *Section,
	const char *Key, 
	const double &Value);

    // It's not legal to overloaded functions by return type;
    // COptions<DString>::Get is called most often, so it's the
    // only one with a short name
    inline DString Get(const char *Key) const;

    virtual DString GetKey(const char *Section, const char *Key) const;

    char *UnknownKeyError(const char *Key);
    int GetComposedMode() { return ComposedKeys; }
    void SetComposedMode(int ComposedMode) { ComposedKeys = ComposedMode; }

    virtual char *AllocMungedSectionName(char *InSection);
};

inline void CUnivOptions::Add(const char *Key, const char *Value)
{
    Add(Key, DString(Value));
}

inline void CUnivOptions::Add(const char *Key, const DString &Value)
{   
    typedef void (COptions<DString>::*FAdd)(
	const char *, const DString &);

    FAdd ptr = &(COptions<DString>::Add);
    (this->*ptr)(Key, Value); 
}

inline void CUnivOptions::Add(const char *Key,
    COptions<DString>::FCompute Compute)
{ 
    typedef void (COptions<DString>::*FAdd)(
	const char *, COptions<DString>::FCompute);

    FAdd ptr = &(COptions<DString>::Add);
    (this->*ptr)(Key, Compute); 
}

inline void CUnivOptions::Add(const char *Key,
    const CStringListValue &Value)
{
    typedef void (COptions<CStringListValue>::*FAdd)(
	const char *, const CStringListValue &);

    FAdd ptr = &(COptions<CStringListValue>::Add);
    (this->*ptr)(Key, Value); 
}

inline void CUnivOptions::Add(const char *Key,
    COptions<CStringListValue>::FCompute Compute)
{ 
    typedef void (COptions<CStringListValue>::*FAdd)(
	const char *, COptions<CStringListValue>::FCompute);

    FAdd ptr = &(COptions<CStringListValue>::Add);
    (this->*ptr)(Key, Compute); 
}

inline void CUnivOptions::Add(const char *Key, int Value)
{
    typedef void (COptions<int>::*FAdd)(
	const char *, const int &);

    FAdd ptr = &(COptions<int>::Add);
    (this->*ptr)(Key, Value); 
}

inline void CUnivOptions::Add(const char *Key,
    COptions<int>::FCompute Compute)
{ 
    typedef void (COptions<int>::*FAdd)(
	const char *, COptions<int>::FCompute);

    FAdd ptr = &(COptions<int>::Add);
    (this->*ptr)(Key, Compute); 
}

inline void CUnivOptions::Add(const char *Key, double &Value)
{
    typedef void (COptions<double>::*FAdd)(
	const char *, const double &);

    FAdd ptr = &(COptions<double>::Add);
    (this->*ptr)(Key, Value); 
}

inline void CUnivOptions::Add(const char *Key,
    COptions<double>::FCompute Compute)
{ 
    typedef void (COptions<double>::*FAdd)(
	const char *, COptions<double>::FCompute);

    FAdd ptr = &(COptions<double>::Add);
    (this->*ptr)(Key, Compute); 
}

inline short CUnivOptions::Override(const char *Section,
    const char *Key, 
    const DString &Value)
{	
    return COptions<DString>::Override(Section, Key, Value); 
}

inline short CUnivOptions::Override(const char *Section,
    const char *Key, 
    const CStringListValue &Value)
{	
    return COptions<CStringListValue>::Override(Section, Key,
	Value); 
}

inline short CUnivOptions::Override(const char *Section,
    const char *Key, 
    int Value)
{ 
    return COptions<int>::Override(Section, Key, Value); 
}

inline short CUnivOptions::Override(const char *Section,
    const char *Key, 
    const double &Value)
{ 
    return COptions<double>::Override(Section, Key, Value); 
}

inline DString CUnivOptions::Get(const char *Key) const
{ 
    return (this->*(&COptions<DString>::Get))(Key); 
}

#endif

